/* mysql_output.c - mysql output driver
 *
 * Copyright (C) 2006-2007 Abilis
 * Copyright (C) 2006-2007 Antek srl
 *
 * Author: Elisa Manara <e@entropika.net>
 * Last update: Mon Nov 20 10:46:02 CET 2006
 *
 * This file is part of Abilis Collector for *nix.
 *
 * Abilis Collector is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * Abilis Collector is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 *                                   
 * You should have received a copy of the GNU General Public License along                        
 * with Abilis Collector; see the file COPYING.  If not, write to the Free                        
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. 
 */  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "define.h"

int checkDBVersion(MYSQL *c) {
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (mysql_query(c, "SELECT * FROM " DATABASE_VERSION_TABLE))
    return (-1);
  result = mysql_store_result(c);
  row = mysql_fetch_row(result);
  if (atoi(row[0]) != DATABASE_VERSION)
    return (-1);

  printDebug("[MySQL] Database version %i found\n", atoi(row[0]));
  return (0);
}

/* Initialize a new MYSQL object and open a connection to MySQL
 * server, considering configuration parameters.
 * Return a pointer to MYSQL object on success, NULL otherwise
 */
int mysqlStart(struct CfgData *config)
{

    if(!mysql_init(&config->mysqlLink)) {
	printError("[MySQL] Cannot allocate a new MySQL object: %s\n", 
	    mysql_error(&config->mysqlLink));
	return -1;
    }

    if (!mysql_real_connect(&config->mysqlLink, config->mysqlHost, 
	config->mysqlUser, config->mysqlPass, config->mysqlDb, 
	config->mysqlPort, NULL, CLIENT_COMPRESS)) {
	    printError("[MySQL] Failed to connect to MySQL database '%s': %s\n",
	    config->mysqlDb, mysql_error(&config->mysqlLink));
	    return -1;
    }   

    /* set automatic reconnection */
    config->mysqlLink.reconnect = 1;

    printDebug("[MySQL] Session opened on server %s\n", config->mysqlHost);

    if (checkDBVersion(&config->mysqlLink) < 0) {
	printError("[MySQL] Database '%s' not updated."
	    " Please see README for how to upgrade it.\n",
	    config->mysqlDb);
	return (-1);
    }

    return 0;
}

int mysqlStop(struct CfgData *config) {
    printDebug("[MySQL] Closing connection...\n");
    mysql_close(&config->mysqlLink);

    return (0);
}

/* send query to MySQL server;
 * return values are:
 *  0  - all right
 *  1  - lost connection
 *  2  - duplicate query
 * -1  - query malformed or error unknown
 */
int mysqlPut(struct CfgData *config, char *query)
{
    int retval;
    unsigned int mysql_err = 0;
    const char* mysql_err_string = NULL;

    pthread_mutex_lock (&config->mysqlMutex);
    retval = mysql_real_query(&config->mysqlLink, query, strlen(query));
    if (retval != 0) {
      mysql_err = mysql_errno(&config->mysqlLink);
      mysql_err_string = mysql_error(&config->mysqlLink);
    }
    pthread_mutex_unlock (&config->mysqlMutex);

    /* successfull query */
    if(retval == 0) return 0;

    /* some error occurs */
    switch(mysql_err) {
    case CR_CONNECTION_ERROR:
    case CR_CONN_HOST_ERROR:
    case CR_UNKNOWN_HOST:
    case CR_SERVER_GONE_ERROR:
    case CR_TCP_CONNECTION:
    case CR_SERVER_LOST:
	/* lost connection */
	printError("[MySQL] Query failed: %s\n", mysql_err_string);
	return 1;
	break;
    case ER_DUP_ENTRY:
	/* duplicate query (tipically reprocessed from queue after a crash */
	printError("[MySQL] Query failed: duplicate key\n");
	return 2;
    default:
	/* other errors, cannot cache query */
	printError("[MySQL] Query rejected: %s\n", mysql_err_string);
	return -1;
	break;
    }
}

/* 
 * vim: softtabstop=4 shiftwidth=4 
 */
